from __future__ import absolute_import
from flask import url_for, render_template, redirect, abort, flash, request,\
        Blueprint, current_app, jsonify
from flask.ext.login import login_required, current_user
from flask.ext.wtf import Form
import six
from six.moves import map
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from wtforms.fields import StringField, SubmitField, HiddenField, SelectField,\
        Label
from wtforms.validators import InputRequired, AnyOf, NumberRange

from ..models import db
from ..auth import PermissionType
from ..auth.models import Division, Permission, Entity


blueprint = Blueprint('divisions', __name__)


@blueprint.route('/')
@login_required
def list_divisions():
    """Show a page listing all divisions.

    Accesible only to administrators.
    """
    if current_user.admin:
        return render_template('divisions.html',
                divisions=Division.query.all())
    if current_user.has_permission(PermissionType.admin):
        admin_permissions = current_user.permissions.filter_by(
                permission=PermissionType.admin).values(Permission.division_id)
        admin_permissions = map(lambda x: x[0], admin_permissions)
        divisions = db.session.query(Division).\
                filter(Division.id.in_(admin_permissions))
        return render_template('divisions.html', divisions=divisions)
    return abort(403)


class AddDivisionForm(Form):
    name = StringField(u'Division Name', validators=[InputRequired()])
    submit = SubmitField(u'Create Division')


@blueprint.route('/add/', methods=['GET', 'POST'])
@login_required
def add_division():
    """Present a form for adding a division and also process that form.

    Only accesible to adminstrators.
    """
    if not current_user.admin:
        return abort(403)
    form = AddDivisionForm()
    if form.validate_on_submit():
        division = Division(form.name.data)
        db.session.add(division)
        db.session.commit()
        return redirect(url_for('.division_detail', division_id=division.id))
    return render_template('form.html', form=form)


class ChangeEntity(Form):
    form_id = HiddenField(default='entity')
    id_ = HiddenField()
    name = StringField()
    permission = HiddenField(validators=[AnyOf(list(PermissionType.values()))])
    action = HiddenField(validators=[AnyOf(('add', 'delete'))])


#: List of tuples enumerating attributes that can be transformed/linked.
#: Mainly used as the choices argument to :py:class:`~.SelectField`
transformer_choices = [
    ('', u''),
    ('kill_timestamp', u'Kill Timestamp'),
    ('pilot', u'Pilot'),
    ('corporation', u'Corporation'),
    ('alliance', u'Alliance'),
    ('system', u'Solar System'),
    ('constellation', u'Constellation'),
    ('region', u'Region'),
    ('ship_type', u'Ship'),
    ('payout', u'Payout'),
    ('status', u'Request Status'),
]


class ChangeTransformer(Form):
    form_id = HiddenField(default='transformer')
    attribute = SelectField(u'Attribute', choices=transformer_choices)
    transformer = SelectField(u'Transformer', choices=[])


def transformer_choices(attr):
    """Conveniece function for generating a list of transformer option tuples.

    :param attr str: the name of the attribute to make a list for.
    :return: A list of tuples suitable for the choices argument of\
        :py:class:`StringField`
    :rtype: list
    """
    default_transformers = [
        ('none', 'None'),
    ]
    choices = default_transformers
    if attr in current_app.url_transformers:
        for transformer in current_app.url_transformers[attr]:
            choices.append((transformer, transformer))
    return choices


@blueprint.route('/<int:division_id>/', methods=['GET', 'POST'])
@login_required
def division_detail(division_id):
    """Generate a page showing the details of a division.

    Shows which groups and individuals have been granted permissions to each
    division.

    Only accesible to administrators.

    :param int division_id: The ID number of the division
    """
    division = Division.query.get_or_404(division_id)
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin, division):
        abort(403)
    if request.method == 'POST':
        if request.form['form_id'] == 'entity':
            form = ChangeEntity()
            if form.validate():
                if form.id_.data != '':
                    entity = Entity.query.get(form.id_.data)
                    if entity is None:
                        flash(u"No entity with ID #{}.".format(form.id_.data),
                                u'error')
                else:
                    try:
                        entity = Entity.query.filter_by(
                                name=form.name.data).one()
                    except NoResultFound:
                        flash(u"No entities with the name '{}' found.".
                                format(form.name.data), category=u'error')
                    except MultipleResultsFound:
                        flash(u"Multiple entities eith the name '{}' found.".
                                format(form.name.data), category=u'error')
                permission_type = PermissionType.from_string(
                        form.permission.data)
                permission_query = Permission.query.filter_by(
                        division=division,
                        entity=entity,
                        permission=permission_type)
                try:
                    permission = permission_query.one()
                except NoResultFound:
                    if form.action.data == 'add':
                        db.session.add(
                            Permission(division, permission_type, entity))
                        flash(u"'{}' is now a {}.".format(entity,
                                permission_type.description.lower()), u"info")
                    elif form.action.data == 'delete':
                        flash(u"{} is not a {}.".format(entity,
                            permission_type.description.lower()), u"warning")
                else:
                    if form.action.data == 'delete':
                        permission_query.delete()
                        flash(u"'{}' is no longer a {}.".format(entity,
                                permission_type.description.lower()), u"info")
                    elif form.action.data == 'add':
                        flash(u"'{}' is now a {}.".format(entity,
                                permission_type.description.lower()), u"info")
            else:
                abort(400)
        elif request.form['form_id'] == 'transformer':
            form = ChangeTransformer()
            attr = form.attribute.data
            form.transformer.choices = transformer_choices(attr)
            # Check form and go from there
            if form.validate():
                name = form.transformer.data
                if name == 'none':
                    division.transformers[attr] = None
                else:
                    # Get the specific map of transformers for the attribute
                    attr_transformers = current_app.url_transformers[attr]
                    # Get the new transformer
                    division.transformers[attr] = attr_transformers[name]
                    # Explicitly add the TransformerRef to the session
                    db.session.add(division.division_transformers[attr])
            else:
                abort(400)
        db.session.commit()
    return render_template(
            'division_detail.html',
            division=division,
            entity_form=ChangeEntity(),
            transformer_form=ChangeTransformer(),
    )


@blueprint.route('/<int:division_id>/transformers/')
@blueprint.route('/<int:division_id>/transformers/<attribute>/')
@login_required
def list_transformers(division_id, attribute=None):
    """API method to get a list of transformers for a division.

    :param division_id int: the ID of the division to look up
    :param attribute str: a specific attribute to look up. Optional.
    :return: JSON
    """
    division = Division.query.get_or_404(division_id)
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin, division):
        abort(403)
    if attribute is None:
        attrs = six.iterkeys(current_app.url_transformers)
    else:
        attrs = (attribute,)
    choices = {}
    for attr in attrs:
        raw_choices = transformer_choices(attr)
        current = division.transformers.get(attr, None)
        if current is not None:
            choices[attr] = \
                    [(c[0], c[1], c[1] == current.name) for c in raw_choices]
        else:
            choices[attr] = \
                    [(c[0], c[1], False) for c in raw_choices]
    return jsonify(choices)



#!/usr/bin.env python

from __future__ import absolute_import
from __future__ import print_function
import os
import os.path
import argparse
import flask
from flask.ext import script
from flask.ext.migrate import Migrate, MigrateCommand, _get_config, stamp
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
from .. import create_app, db, migrate


manager = script.Manager(create_app)


class AbsolutePathAction(argparse.Action):
    """Custom argparse.Action that transforms path strings into absolute paths.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        current_absolute = os.path.abspath(os.getcwd())
        if isinstance(values, str):
            new_values = os.path.join(current_absolute, values)
        else:
            new_values = []
            for value in values:
                real_path = os.path.join(current_absolute, values)
                new_values.append(real_path)
        setattr(namespace, self.dest, new_values)


# Monkeypatch Flask-Script to consider my custom path action 'safe'
safe_actions = list(script.safe_actions)
safe_actions.append(AbsolutePathAction)
script.safe_actions = safe_actions


# Crufty workaround to get Flask-Migrate working with app factories
class MigrateManager(script.Manager):
    def __init__(self, old_command, directory='migrations'):
        super(MigrateManager, self).__init__()
        self.directory = directory
        for attr in ('app', '_commands', '_options', 'usage', 'help',
                'description', 'disable_argcomplete', 'with_default_commands',
                'parent'):
            setattr(self, attr, getattr(old_command, attr))

    def __call__(self, app=None, directory='migrations', **kwargs):
        if app is None:
            app = self.app
            if app is None:
                raise Exception(u"No app specified")

        db = kwargs.pop('db', None)

        if not isinstance(app, flask.Flask):
            app = app(**kwargs)

        # Last ditch effort to get a database handle
        if db is None:
            if 'sqlalchemy' in app.extensions:
                db = app.extensions['sqlalchemy'].db
            else:
                raise Exception(u"No database defined for app.")

        Migrate(app, db, self.directory)
        return app


migrate_path = os.path.dirname(migrate.__file__)
migrate_path = os.path.abspath(migrate_path)
migrate_manager = MigrateManager(MigrateCommand, migrate_path)
manager.add_command('db', migrate_manager)


manager.add_option('-c', '--config', dest='config', required=True,
        action=AbsolutePathAction)


@migrate_manager.command
def create(force=False):
    """Create tables if the database has not been configured yet."""
    # Fail if there's an alembic version set
    engine = db.get_engine(flask.current_app)
    conn = engine.connect()
    context = MigrationContext.configure(conn)
    current_rev = context.get_current_revision()
    alembic_config = _get_config(directory=migrate_path)
    script = ScriptDirectory.from_config(alembic_config)
    latest_rev = script.get_current_head()
    if current_rev == latest_rev and not force:
        print(u"You need to run 'evesrp -c config.py db migrate' to "
              u"migrate to the latest database schema.")
    else:
        db.create_all()
        if current_rev is None:
            stamp()


def main():
    manager.run()


if __name__ == '__main__':
    main()



from __future__ import absolute_import
from ecdsa import SigningKey, VerifyingKey, NIST256p
from brave.api.client import SignedAuth, API
from sqlalchemy.orm.exc import NoResultFound
from flask import flash, url_for, redirect, abort, current_app, request
from hashlib import sha256
from binascii import unhexlify

from .. import db, requests_session
from . import AuthMethod, AuthForm
from .models import User, Group, Pilot


class BraveCore(AuthMethod):
    def __init__(self, client_key, server_key, identifier,
            url='https://core.braveineve.com', **kwargs):
        self.api = API(url, identifier, client_key, server_key,
                requests_session).api
        if 'name' not in kwargs:
            kwargs['name'] = u'Brave Core'
        super(BraveCore, self).__init__(**kwargs)

    def login(self, form):
        # Redirect to Core for the authorization token. Give URLs to return to.
        # Sidenote: At this time, Brave has nearly 0 API documentation. The
        # kinda-sorta hidden TEST Auth API documentation is more complete.
        result_url = url_for('login.auth_method_login', _external=True,
                auth_method=self.safe_name)
        response = self.api.core.authorize(success=result_url,
                failure=result_url)
        core_url = response[u'location']
        return redirect(core_url)

    def list_groups(self, user=None):
        pass

    def view(self):
        token = unistr.ensure_unicode(request.args.get('token'))
        if token is not None:
            info = self.api.core.info(token=token)
            char_name = info.character.name
            try:
                user = CoreUser.query.filter_by(name=char_name,
                        authmethod=self.name).one()
                user.token = token
            except NoResultFound:
                user = CoreUser(name=char_name, authmethod=self.name,
                        token=token)
                db.session.add(user)
            # Apply admin flag
            user.admin = user.name in self.admins
            # Sync up group membership
            for group_name in info.tags:
                try:
                    group = CoreGroup.query.filter_by(name=group_name,
                            authmethod=self.name).one()
                except NoResultFound:
                    group = CoreGroup(group_name, self.name)
                    db.session.add(group)
                user.groups.add(group)
            for group in user.groups:
                if group.name not in info.tags:
                    user.groups.remove(group)
            # Sync pilot (just the primary for now)
            pilot = Pilot.query.get(info.character.id)
            if not pilot:
                pilot = Pilot(user, char_name, info.character.id)
                db.session.add(pilot)
            else:
                pilot.user = user
            db.session.commit()
            self.login_user(user)
            # TODO Have a meaningful redirect for this
            return redirect(url_for('index'))
        else:
            flash(u"Login failed.", u'error')
            return redirect(url_for('login.login'))


class CoreUser(User):
    id = db.Column(db.Integer, db.ForeignKey('user.id'), primary_key=True)
    token = db.Column(db.String(100, convert_unicode=True))


class CoreGroup(Group):
    id = db.Column(db.Integer, db.ForeignKey('group.id'), primary_key=True)
    description = db.Column(db.Text(convert_unicode=True))



from __future__ import absolute_import
from collections import defaultdict
import datetime as dt
import time
from decimal import Decimal
from functools import partial
import re
import sys
import six
from .util.unistr import unistr
from .util.urlparse import urlparse, urlunparse
from .util.utc import utc

import requests
from sqlalchemy import create_engine, Table, MetaData
from sqlalchemy.sql import select

from . import ships, systems


if six.PY3:
    unicode = str


@unistr
class Killmail(object):
    """Base killmail representation.

    .. py:attribute:: kill_id

        The ID integer of this killmail. Used by most killboards and by CCP to
        refer to killmails.

    .. py:attribute:: ship_id

        The typeID integer of for the ship lost for this killmail.

    .. py:attribute:: ship

        The human readable name of the ship lost for this killmail.

    .. py:attribute:: ship_url

        This is an optional atribute for subclasses to implement. It's intended
        to be used for requests to link to a custom, possibly external,
        ship-specific page.

    .. py:attribute:: pilot_id

        The ID number of the pilot who lost the ship. Referred to by CCP as
        ``characterID``.

    .. py:attribute:: pilot

        The name of the pilot who lost the ship.

    .. py:attribute:: corp_id

        The ID number of the corporation :py:attr:`pilot` belonged to at the
        time this kill happened.

    .. py:attribute:: corp

        The name of the corporation referred to by :py:attr:`corp_id`.

    .. py:attribute:: alliance_id

        The ID number of the alliance :py:attr:`corp` belonged to at the time
        of this kill, or ``None`` if the corporation wasn't in an alliance at
        the time.

    .. py:attribute:: alliance

        The name of the alliance referred to by :py:attr:`alliance_id`.

    .. py:attribute:: url

        A URL for viewing this killmail's information later. Typically an
        online killboard such as `zKillboard <https://zkillboard.com>`, but
        other kinds of links may be used.

    .. py:attribute:: value

        The extimated ISK loss for the ship destroyed in this killmail. This is
        an optional attribute, and is ``None`` if unsupported. If this
        attribute is set, it should be a :py:class:`~.Decimal` or a type that
        can be used as the value for the Decimal constructor.

    .. py:attribute:: timestamp

        The date and time that this kill occured as a
        :py:class:`datetime.datetime` object (with a UTC timezone).

    .. py:attribute:: verified

        Whether or not this killmail has been API verified (or more accurately,
        if it is to be trusted when making a
        :py:class:`~evesrp.models.Request`.

    .. py:attribute:: system
    .. py:attribute:: system_id
    .. py:attribute:: constellation
    .. py:attribute:: region

        The system/constellation/region where the kill occured.
    """

    def __init__(self, **kwargs):
        """Initialize a :py:class:`Killmail` with ``None`` for all attributes.

        All subclasses of this class, (and all mixins designed to be used with
        it) must call ``super().__init__(**kwargs)`` to ensure all
        initialization is done.

        :param: keyword arguments corresponding to attributes.
        """
        self._data = defaultdict(lambda: None)
        for attr in (u'kill_id', u'ship_id', u'ship', u'pilot_id', u'pilot',
                u'corp_id', u'corp', u'alliance_id', u'alliance', u'verified',
                u'url', u'value', u'timestamp', u'system', u'constellation',
                u'region', u'system_id'):
            try:
                setattr(self, attr, kwargs[attr])
            except KeyError:
                pass
            else:
                del kwargs[attr]
        super(Killmail, self).__init__(**kwargs)

    # Any attribute not starting with an underscore will now be stored in a
    # separate, private attribute. This is to allow attribute on Killmail to be
    # redfined as a property.

    def __getattr__(self, name):
        try:
            return self._data[name]
        except KeyError as e:
            raise AttributeError(unicode(e))

    def __setattr__(self, name, value):
        if name[0] == '_':
            object.__setattr__(self, name, value)
        else:
            self._data[name] = value

    def __unicode__(self):
        return u"{kill_id}: {pilot} lost a {ship}. Verified: {verified}.".\
                format(kill_id=self.kill_id, pilot=self.pilot, ship=self.ship,
                        verified=self.verified)

    def __iter__(self):
        """Iterate over the attributes of this killmail.

        Yields tuples in the form ``('<name>', <value>)``. This is used by
        :py:meth:`evesrp.models.Request.__init__` to initialize its data
        quickly. The `<name>` in the returned tuples is the name of the
        attribute on the :py:class:`~evesrp.models.Request`.
        """
        yield ('id', self.kill_id)
        yield ('ship_type', self.ship)
        yield ('corporation', self.corp)
        yield ('alliance', self.alliance)
        yield ('killmail_url', self.url)
        yield ('base_payout', self.value)
        yield ('kill_timestamp', self.timestamp)
        yield ('system', self.system)
        yield ('constellation', self.constellation)
        yield ('region', self.region)


class ShipNameMixin(object):
    """Killmail mixin providing :py:attr:`Killmail.ship` from
    :py:attr:`Killmail.ship_id`.
    """

    @property
    def ship(self):
        """Looks up the ship name using :py:attr:`Killmail.ship_id`.
        """
        return ships.ships[self.ship_id]


class LocationMixin(object):
    """Killmail mixin for providing solar system, constellation and region
    names from :py:attr:`Killmail.system_id`.
    """

    @property
    def system(self):
        """Provides the solar system name using :py:attr:`Killmail.system_id`.
        """
        return systems.system_names[self.system_id]

    @property
    def constellation(self):
        """Provides the constellation name using :py:attr:`Killmail.system_id`.
        """
        return systems.systems_constellations[self.system]

    @property
    def region(self):
        """Provides the region name using :py:attr:`Killmail.system_id`.
        """
        return systems.constellations_regions[self.constellation]


class RequestsSessionMixin(object):
    """Mixin for providing a :py:class:`requests.Session`.

    The shared session allows HTTP user agents to be set properly, and for
    possible connection pooling.

    .. py:attribute:: requests_session

        A :py:class:`~requests.Session` for making HTTP requests.
    """
    def __init__(self, requests_session=None, **kwargs):
        """Set up a :py:class:`~requests.Session` for making HTTP requests.

        If an existing session is not provided, one will be created.

        :param requests_session: an existing session to use.
        :type requests: :py:class:`~requests.Session`
        """
        if requests_session is None:
            self.requests_session = requests.Session()
        else:
            self.requests_session = requests_session
        super(RequestsSessionMixin, self).__init__(**kwargs)


class ZKillmail(Killmail, RequestsSessionMixin, ShipNameMixin, LocationMixin):
    """A killmail sourced from a zKillboard based killboard."""

    zkb_regex = re.compile(r'/(detail|kill)/(?P<kill_id>\d+)/?')

    def __init__(self, url, **kwargs):
        """Create a killmail from the given URL.

        :param str url: The URL of the killmail.
        :raises ValueError: if ``url`` isn't a valid zKillboard killmail.
        :raises LookupError: if the zKillboard API response is in an unexpected
            format.
        """
        super(ZKillmail, self).__init__(**kwargs)
        self.url = url
        match = self.zkb_regex.search(url)
        if match:
            self.kill_id = int(match.group('kill_id'))
        else:
            raise ValueError(u"'{}' is not a valid zKillboard killmail".
                    format(self.url))
        parsed = urlparse(self.url, scheme='https')
        if parsed.netloc == '':
            # Just in case someone is silly and gives an address without a
            # scheme. Also fix self.url to have a scheme.
            parsed = urlparse('//' + url, scheme='https')
            self.url = parsed.geturl()
        self.domain = parsed.netloc
        # Check API
        api_url = [a for a in parsed]
        api_url[2] = '/api/no-attackers/no-items/killID/{}'.format(
                self.kill_id)
        resp = self.requests_session.get(urlunparse(api_url))
        retrieval_error = LookupError(u"Error retrieving killmail data: {}"
                    .format(resp.status_code))
        if resp.status_code != 200:
            raise retrieval_error
        try:
            json = resp.json()
        except ValueError as e:
            raise retrieval_error
        try:
            json = json[0]
        except IndexError as e:
            raise LookupError(u"Invalid killmail: {}".format(url))
        # JSON is defined to be UTF-8 in the standard
        victim = json[u'victim']
        self.pilot_id = int(victim[u'characterID'])
        self.pilot = victim[u'characterName']
        self.corp_id = int(victim[u'corporationID'])
        self.corp = victim[u'corporationName']
        if victim[u'allianceID'] != '0':
            self.alliance_id = int(victim[u'allianceID'])
            self.alliance = victim[u'allianceName']
        self.ship_id = int(victim[u'shipTypeID'])
        self.system_id = int(json[u'solarSystemID'])
        # For consistency, store self.value in millions. Decimal is being used
        # for precision at large values.
        # Old versions of zKB don't give the ISK value
        try:
            self.value = Decimal(json[u'zkb'][u'totalValue'])
        except KeyError:
            self.value = Decimal(0)
        # Parse the timestamp
        time_struct = time.strptime(json[u'killTime'], '%Y-%m-%d %H:%M:%S')
        self.timestamp = dt.datetime(*(time_struct[0:6]),
                tzinfo=utc)

    @property
    def verified(self):
        # zKillboard assigns unverified IDs negative numbers
        return self.kill_id > 0

    def __unicode__(self):
        parent = super(ZKillmail, self).__unicode__()
        return u"{parent} From ZKillboard at {url}".format(parent=parent,
                url=self.url)


class CRESTMail(Killmail, RequestsSessionMixin, LocationMixin):
    """A killmail with data sourced from a CREST killmail link."""

    crest_regex = re.compile(r'/killmails/(?P<kill_id>\d+)/[0-9a-f]+/')

    def __init__(self, url, **kwargs):
        """Create a killmail from a CREST killmail link.

        :param str url: the CREST killmail URL.
        :raises ValueError: if ``url`` is not a CREST URL.
        :raises LookupError: if the CREST API response is in an unexpected
            format.
        """
        super(CRESTMail, self).__init__(**kwargs)
        self.url = url
        match = self.crest_regex.search(self.url)
        if match:
            self.kill_id = match.group('kill_id')
        else:
            raise ValueError(u"'{}' is not a valid CREST killmail".
                    format(self.url))
        parsed = urlparse(self.url, scheme='https')
        if parsed.netloc == '':
            parsed = urlparse('//' + url, scheme='https')
            self.url = parsed.geturl()
        # Check if it's a valid CREST URL
        resp = self.requests_session.get(self.url)
        # JSON responses are defined to be UTF-8 encoded
        if resp.status_code != 200:
            raise LookupError(u"Error retrieving CREST killmail: {}".format(
                resp.json()[u'message']))
        try:
            json = resp.json()
        except ValueError as e:
            raise LookupError(u"Error retrieving killmail data: {}"
                    .format(resp.status_code))
        victim = json[u'victim']
        char = victim[u'character']
        corp = victim[u'corporation']
        ship = victim[u'shipType']
        alliance = victim[u'alliance']
        self.pilot_id = char[u'id']
        self.pilot = char[u'name']
        self.corp_id = corp[u'id']
        self.corp = corp[u'name']
        self.alliance_id = alliance[u'id']
        self.alliance = alliance[u'name']
        self.ship_id = ship[u'id']
        self.ship = ship[u'name']
        solarSystem = json[u'solarSystem']
        self.system_id = solarSystem[u'id']
        self.system = solarSystem[u'name']
        # CREST Killmails are always verified
        self.verified = True
        # Parse the timestamp
        time_struct = time.strptime(json[u'killTime'], '%Y.%m.%d %H:%M:%S')
        self.timestamp = dt.datetime(*(time_struct[0:6]),
                tzinfo=utc)



from datetime import tzinfo, timedelta


class UTC(tzinfo):

    def utcoffset(self, dt):
        return timedelta()

    def dst(self, dt):
        return timedelta()

    def tzname(self, dt):
        return u'UTC+00:00'


utc = UTC()



from __future__ import absolute_import
from flask import url_for
from flask.json import JSONEncoder
import six
from .models import Request, Action, Modifier
from .auth.models import User, Group, Division, APIKey

if six.PY3:
    unicode = str


class IterableEncoder(JSONEncoder):
    def default(self, o):
        try:
            iterator = iter(o)
        except TypeError:
            pass
        else:
            return list(o)
        return super(IterableEncoder, self).default(o)


class NamedEncoder(JSONEncoder):
    def default(self, o):
        try:
            ret = {
                u'name': o.name,
                u'id': o.id,
            }
        except AttributeError:
            pass
        else:
            if isinstance(o, User):
                ret[u'href'] = url_for('api.user_detail', user_id=o.id)
            elif isinstance(o, Group):
                ret[u'href'] = url_for('api.group_detail', group_id=o.id)
            elif isinstance(o, Division):
                ret[u'href'] = url_for('api.division_detail', division_id=o.id)
            if u'href' in ret:
                return ret
        return super(NamedEncoder, self).default(o)


class GrabbagEncoder(JSONEncoder):
    def default(self, o):
        try:
            ret = {'id': o.id}
        except AttributeError:
            pass
        else:
            if isinstance(o, Request):
                ret[u'href'] = url_for('requests.get_request_details',
                        request_id=o.id)
                attrs = (u'killmail_url', u'kill_timestamp', u'pilot',
                         u'alliance', u'corporation', u'submitter',
                         u'division', u'status', u'base_payout', u'payout',
                         u'details', u'id')
                for attr in attrs:
                    if attr == u'pilot':
                        ret[attr] = str(o.pilot)
                    elif attr == u'status':
                        ret[attr] = o.status.value
                    elif u'payout' in attr:
                        payout = getattr(o, attr)
                        ret[attr] = payout.currency(commas=False)
                        ret[attr + '_str'] = payout.currency()
                    else:
                        ret[attr] = getattr(o, attr)
                ret[u'submit_timestamp'] = o.timestamp
                return ret
            elif isinstance(o, Action):
                ret[u'note'] = o.note or u''
                ret[u'timestamp'] = o.timestamp
                ret[u'user'] = o.user
                ret[u'type'] = o.type_.value
                return ret
            elif isinstance(o, Modifier):
                ret[u'user'] = o.user
                ret[u'timestamp'] = o.timestamp
                ret[u'note'] = o.note or u''
                if o.voided:
                    ret[u'void'] = {
                            u'user': o.voided_user,
                            u'timestamp': o.voided_timestamp,
                    }
                else:
                    ret[u'void'] = False
                if hasattr(o.value, 'currency'):
                    ret[u'value'] = o.value.currency(commas=False)
                else:
                    ret[u'value'] = o.value
                ret[u'value_str'] = unicode(o)
                return ret
            elif isinstance(o, APIKey):
                ret[u'key'] = o.hex_key
                ret[u'timestamp'] = o.timestamp
                return ret
        return super(GrabbagEncoder, self).default(o)


# Multiple inheritance FTW
class SRPEncoder(IterableEncoder, NamedEncoder, GrabbagEncoder):
    pass



from __future__ import absolute_import
import datetime as dt
from decimal import Decimal
import locale
import six
from six.moves import filter, map, range
from sqlalchemy.types import DateTime
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from flask import Markup

from . import db
from .util import DeclEnum, classproperty, AutoID, Timestamped, AutoName,\
        unistr
from .auth import PermissionType


if six.PY3:
    unicode = str


class PrettyDecimal(Decimal):
    """:py:class:`~.Decimal` subclass that can pretty-print itself."""

    def currency(self, commas=True):
        """Format the Decimal as a currency number.

        Commas for the thousands separators, full stops for the decimal, two
        decimal places. Commas are optional and controlled by the ``commas``
        argument.

        Adapted from https://docs.python.org/3.3/library/decimal.html#recipes
        """
        sign, digits, exp = self.quantize(Decimal('0.01')).as_tuple()
        digits = list(map(unicode, digits))
        result = []
        for i in range(2):
            result.append(digits.pop() if digits else u'0')
        result.append(u'.')
        if not digits:
            result.append(u'0')
        count = 0
        while digits:
            result.append(digits.pop())
            count += 1
            if count == 3 and digits and commas:
                count = 0
                result.append(u',')
        result.append(u'-' if sign else u'')
        return u''.join(reversed(result))


class PrettyNumeric(db.TypeDecorator):
    """Type Decorator for :py:class:`~.Numeric` that reformats the userland
    values into :py:class:`PrettyDecimal`\s.
    """

    impl = db.Numeric

    def process_result_value(self, value, dialect):
        return PrettyDecimal(value) if value is not None else None


class ActionType(DeclEnum):

    # The actual stored values are single character to make it easier on
    # engines that don't support native enum types.

    #: Status for a request being evaluated.
    evaluating = u'evaluating', u'Evaluating'

    #: Status for a request that has been evaluated and is awaitng payment.
    approved = u'approved', u'Approved'

    #: Status for a request that has been paid. This is a terminatint state.
    paid = u'paid', u'Paid'

    #: Status for a requests that has been rejected. This is a terminating
    #: state.
    rejected = u'rejected', u'Rejected'

    #: Status for a request that is missing details and needs further action.
    incomplete = u'incomplete', u'Incomplete'

    #: A special type of :py:class:`Action` representing a comment made on the
    #: request.
    comment = u'comment', u'Comment'

    @classproperty
    def finalized(cls):
        return frozenset((cls.paid, cls.rejected))

    @classproperty
    def pending(cls):
        return frozenset((cls.evaluating, cls.approved, cls.incomplete))

    @classproperty
    def statuses(cls):
        return frozenset((cls.evaluating, cls.approved, cls.paid, cls.rejected,
                cls.incomplete))


class ActionError(ValueError):
    """Error raised for invalid state changes for a :py:class:`Request`."""
    pass


class ModifierError(ValueError):
    """Error raised when a modification is attempted to a :py:class:`Request`
    when it's in an invalid state.
    """
    pass


class Action(db.Model, AutoID, Timestamped, AutoName):
    """Actions change the state of a Request.
    
    With the exception of the comment action (which does nothing), actions
    change the state of a Request.
    """

    #: The action be taken. See :py:class:`ActionType` for possible values.
    type_ = db.Column(ActionType.db_type(), nullable=False)

    #: The ID of the :py:class:`Request` this action applies to.
    request_id = db.Column(db.Integer, db.ForeignKey('request.id'))

    #: The :py:class:`Request` this action applies to.
    request = db.relationship('Request', back_populates='actions')

    #: The ID of the :py:class:`~.User` who made this action.
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    #: The :py:class:`~.User` who made this action.
    user = db.relationship('User', back_populates='actions')

    #: Any additional notes for this action.
    note = db.Column(db.Text(convert_unicode=True))

    def __init__(self, request, user, note=None, type_=None):
        if type_ is not None:
            self.type_ = type_
        self.user = user
        self.note = unistr.ensure_unicode(note)
        self.timestamp = dt.datetime.utcnow()
        self.request = request

    @db.validates('type_')
    def set_request_type(self, attr, type_):
        if self.request is not None:
            if type_ != ActionType.comment and self.timestamp >=\
                    self.request.actions[0].timestamp:
                self.request.status = type_
        return type_

    def __repr__(self):
        return "{x.__class__.__name__}({x.request}, {x.user}, {x.type_})".\
                format(x=self)


class Modifier(db.Model, AutoID, Timestamped, AutoName):
    """Modifiers apply bonuses or penalties to Requests.

    This is an abstract base class for the pair of concrete implementations.
    Modifiers can be voided at a later date. The user who voided a modifier and
    when they did are recorded.
    """

    #: Discriminator column for SQLAlchemy
    _type = db.Column(db.String(20, convert_unicode=True), nullable=False)

    #: The ID of the :py:class:`Request` this modifier applies to.
    request_id = db.Column(db.Integer, db.ForeignKey('request.id'))

    #: The :py:class:`Request` this modifier applies to.
    request = db.relationship('Request', back_populates='modifiers')

    #: The ID of the :py:class`~.User` who added this modifier.
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)

    #: The :py:class:`~.User` who added this modifier.
    user = db.relationship('User', foreign_keys=[user_id])

    #: Any notes explaining this modification.
    note = db.Column(db.Text(convert_unicode=True))

    #: The ID of the :py:class:`~.User` who voided this modifier (if voided).
    voided_user_id = db.Column(db.Integer, db.ForeignKey('user.id'),
            nullable=True)

    #: The :py:class:`~.User` who voided this modifier if it has been voided.
    voided_user = db.relationship('User', foreign_keys=[voided_user_id])

    #: If this modifier has been voided, this will be the timestamp of when it
    #: was voided.
    voided_timestamp = db.Column(DateTime)

    @hybrid_property
    def voided(self):
        """Boolean of whether this modifier has been voided or not."""
        return self.voided_user is not None and \
                self.voided_timestamp is not None

    @voided.expression
    def voided(cls):
        return db.and_(
                cls.voided_user_id != None,
                cls.voided_timestamp != None
        )

    @declared_attr
    def __mapper_args__(cls):
        """SQLAlchemy late-binding attribute to set mapper arguments.

        Obviates subclasses from having to specify polymorphic identities.
        """
        cls_name = unicode(cls.__name__)
        args = {'polymorphic_identity': cls_name}
        if cls_name == u'Modifier':
            args['polymorphic_on'] = cls._type
        return args

    def __init__(self, request, user, note, value):
        self.user = user
        self.note = unistr.ensure_unicode(note)
        self.value = value
        self.request = request

    def __repr__(self):
        return ("{x.__class__.__name__}({x.request}, {x.user},"
                "{x}, {x.voided})".format(x=self, value=self))

    def void(self, user):
        """Mark this modifier as void.

        :param user: The user voiding this modifier
        :type user: :py:class:`~.User`
        """
        if self.request.status != ActionType.evaluating:
            raise ModifierError("Modifiers can only be voided when the request"
                                " is in the evaluating state.")
        if not user.has_permission(PermissionType.review,
                self.request.division):
            raise ModifierError("You must be a reviewer to be able to void "
                                "modifiers.")
        self.voided_user = user
        self.voided_timestamp = dt.datetime.utcnow()


@unistr.unistr
class AbsoluteModifier(Modifier):
    """Subclass of :py:class:`Modifier` for representing absolute
    modifications.

    Absolute modifications are those that are not dependent on the value of
    :py:attr:`Request.base_payout`.
    """

    id = db.Column(db.Integer, db.ForeignKey('modifier.id'), primary_key=True)

    #: How much ISK to add or remove from the payout
    value = db.Column(PrettyNumeric(precision=15, scale=2), nullable=False, default=0.0)

    def __unicode__(self):
        return u'{}M ISK {}'.format(self.value / 1000000,
                u'bonus' if self.value >= 0 else u'penalty')


@unistr.unistr
class RelativeModifier(Modifier):
    """Subclass of :py:class:`Modifier` for representing relative modifiers.

    Relative modifiers depend on the value of :py:attr:`Modifier.base_payout`
    to calculate their effect.
    """

    id = db.Column(db.Integer, db.ForeignKey('modifier.id'), primary_key=True)

    #: What percentage of the payout to add or remove
    value = db.Column(db.Float, nullable=False, default=0.0)

    def __unicode__(self):
        return u'{}% {}'.format(self.value * 100,
                u'bonus' if self.value >= 0 else u'penalty')


class Request(db.Model, AutoID, Timestamped, AutoName):
    """Requests represent SRP requests."""

    #: The ID of the :py:class:`~.User` who submitted this request.
    submitter_id = db.Column(db.Integer, db.ForeignKey('user.id'))

    #: The :py:class:`~.User` who submitted this request.
    submitter = db.relationship('User', back_populates='requests')

    #: The ID of the :py:class`~.Division` this request was submitted to.
    division_id = db.Column(db.Integer, db.ForeignKey('division.id'),
            nullable=False)

    #: The :py:class:`~.Division` this request was submitted to.
    division = db.relationship('Division', back_populates='requests')

    #: A list of :py:class:`Action`\s that have been applied to this request,
    #: sorted in the order they were applied.
    actions = db.relationship('Action', back_populates='request',
            order_by='desc(Action.timestamp)')

    #: A list of all :py:class:`Modifier`\s that have been applied to this
    #: request, regardless of wether they have been voided or not. They're
    #: sorted in the order they were added.
    modifiers = db.relationship('Modifier', back_populates='request',
            order_by='desc(Modifier.timestamp)', lazy='dynamic')

    #: The URL of the source killmail.
    killmail_url = db.Column(db.String(512, convert_unicode=True),
            nullable=False)

    #: The ID of the :py:class:`~.Pilot` for the killmail.
    pilot_id = db.Column(db.Integer, db.ForeignKey('pilot.id'), nullable=False)

    #: The :py:class:`~.Pilot` for the killmail this request is for.
    pilot = db.relationship('Pilot', back_populates='requests')

    #: The corporation of the :py:attr:`pilot` at the time of the killmail.
    corporation = db.Column(db.String(150, convert_unicode=True),
            nullable=False, index=True)

    #: The alliance of the :py:attr:`pilot` at the time of the killmail.
    alliance = db.Column(db.String(150, convert_unicode=True), nullable=True,
            index=True)

    #: The type of ship that was destroyed.
    ship_type = db.Column(db.String(75, convert_unicode=True), nullable=False,
            index=True)

    # TODO: include timezones
    #: The date and time of when the ship was destroyed.
    kill_timestamp = db.Column(DateTime, nullable=False, index=True)

    #: The base payout for this request in millions of ISK.
    #: :py:attr:`modifiers` apply to this value.
    base_payout = db.Column(PrettyNumeric(precision=15, scale=2), default=0.0)

    #: Supporting information for the request.
    details = db.deferred(db.Column(db.Text(convert_unicode=True)))

    #: The current status of this request
    status = db.Column(ActionType.db_type(), nullable=False,
            default=ActionType.evaluating)

    #: The solar system this loss occured in.
    system = db.Column(db.String(25, convert_unicode=True), nullable=False,
            index=True)

    #: The constellation this loss occured in.
    constellation = db.Column(db.String(25, convert_unicode=True),
            nullable=False, index=True)

    #: The region this loss occured in.
    region = db.Column(db.String(25, convert_unicode=True), nullable=False,
            index=True)

    @property
    def payout(self):
        """The resulting payout taking all active :py:attr:`modifiers` into
        account.

        The return value is an internal class that will return different
        representations depending on the type it is being coerced to.
        :py:class:`Strings <str>` will be formatted accroding to the current
        locale with thousands separators, :py:func:`float`\s will be in
        millions of ISK, and :py:func:`ints`\s will be the total ISK value
        (equivalent to the string representation).
        """
        # Evaluation method for payout:
        # almost_payout = (sum(absolute_modifiers) + base_payout)
        # payout = almost_payout + (sum(percentage_modifiers) * almost_payout)
        abs_mods = db.session.query(db.func.sum(AbsoluteModifier.value))\
                .join(Request)\
                .filter(Modifier.request_id==self.id)\
                .filter(~Modifier.voided)
        rel_mods = db.session.query(db.func.sum(RelativeModifier.value))\
                .join(Request)\
                .filter(Modifier.request_id==self.id)\
                .filter(~Modifier.voided)
        absolute = abs_mods.one()[0]
        if absolute is None:
            absolute = Decimal(0)
        relative = rel_mods.one()[0]
        if relative is None:
            relative = Decimal(0)
        else:
            relative = Decimal.from_float(relative)
        payout = self.base_payout + absolute
        payout = payout + (payout * relative)

        return PrettyDecimal(payout)

    @hybrid_property
    def finalized(self):
        """If this request is in a finalized status (``'paid'`` or
        ``'rejected'``).
        """
        return self.status in ActionType.finalized

    @finalized.expression
    def finalized(cls):
        return db.or_(cls.status == ActionType.paid,
                cls.status == ActionType.rejected)

    def __init__(self, submitter, details, division, killmail, **kwargs):
        """Create a :py:class:`Request`.

        :param submitter: The user submitting this request
        :type submitter: :py:class:`~.User`
        :param str details: Supporting details for this request
        :param division: The division this request is being submitted to
        :type division: :py:class:`~.Division`
        :param killmail: The killmail this request pertains to
        :type killmail: :py:class:`~.Killmail`
        """
        self.division = division
        self.details = details
        self.submitter = submitter
        # Pull basically everything else from the killmail object
        # The base Killmail object has an iterator defined that returns tuples
        # of Request attributes and values for those attributes
        for attr, value in killmail:
            setattr(self, attr, value)
        # Set default values before a flush
        if self.base_payout is None and 'base_payout' not in kwargs:
            self.base_payout = Decimal(0)
        super(Request, self).__init__(**kwargs)

    @db.validates('base_payout')
    def validate_payout(self, attr, value):
        """Ensures that base_payout is positive. The value is clamped to 0."""
        if self.status == ActionType.evaluating or self.status is None:
            if value is None or value < 0:
                return Decimal('0')
            else:
                return Decimal(value)
        else:
            raise ModifierError(u"The request must be in the evaluating state "
                                u"to change the base payout.")

    state_rules = {
        ActionType.evaluating: {
            ActionType.incomplete: (PermissionType.review,),
            ActionType.rejected: (PermissionType.review,),
            ActionType.approved: (PermissionType.review,),
            ActionType.evaluating: (PermissionType.review,
                PermissionType.submit),
        },
        ActionType.incomplete: {
            ActionType.rejected: (PermissionType.review,),
            ActionType.evaluating: (PermissionType.review,
                PermissionType.submit),
        },
        ActionType.rejected: {
            ActionType.evaluating: (PermissionType.review,),
        },
        ActionType.approved: {
            ActionType.evaluating: (PermissionType.review,),
            ActionType.paid: (PermissionType.pay,),
        },
        ActionType.paid: {
            ActionType.approved: (PermissionType.pay,),
            ActionType.evaluating: (PermissionType.pay,),
        },
    }

    def valid_actions(self, user):
        """Get valid actions (besides comment) the given user can perform."""
        possible_actions = self.state_rules[self.status]
        def action_filter(action):
            return user.has_permission(possible_actions[action],
                    self.division)
        return filter(action_filter, possible_actions)

    @db.validates('status')
    def validate_status(self, attr, new_status):
        """Enforces that status changes follow the status state diagram below.
        When an invalid change is attempted, :py:class:`ActionError` is
        raised.

        .. digraph:: request_workflow

            rankdir="LR";

            sub [label="submitted", shape=plaintext];

            node [style="dashed, filled"];

            eval [label="evaluating", fillcolor="#fcf8e3"];
            rej [label="rejected", style="solid, filled", fillcolor="#f2dede"];
            app [label="approved", fillcolor="#d9edf7"];
            inc [label="incomplete", fillcolor="#f2dede"];
            paid [label="paid", style="solid, filled", fillcolor="#dff0d8"];

            sub -> eval;
            eval -> rej [label="R"];
            eval -> app [label="R"];
            eval -> inc [label="R"];
            rej -> eval [label="R"];
            inc -> eval [label="R, S"];
            inc -> rej [label="R"];
            app -> paid [label="P"];
            app -> eval [label="R"];
            paid -> eval [label="P"];
            paid -> app [label="P"];

        R means a reviewer can make that change, S means the submitter can make
        that change, and P means payers can make that change. Solid borders are
        terminal states.
        """

        def check_status(*valid_states):
            if new_status not in valid_states:
                raise ActionError(u"{} is not a valid status to change "
                        u"to from {} (valid options: {})".format(new_status,
                                self.status, valid_states))


        if new_status == ActionType.comment:
            raise ValueError(u"ActionType.comment is not a valid status")
        # Initial status
        if self.status is None:
            return new_status
        rules = self.state_rules[self.status]
        if new_status not in rules:
            raise ActionError(u"{} is not a valid status to change to from {} "
                    u"(valid options: {})".format(new_status,
                            self.status, list(six.iterkeys(rules))))
        return new_status

    @db.validates('actions')
    def update_status_from_action(self, attr, action):
        """Updates :py:attr:`status` whenever a new :py:class:`~.Action`
        is added and verifies permissions.
        """
        if action.type_ is None:
            # Action.type_ are not nullable, so rely on the fact that it will
            # be set later to let it slide now.
            return action
        elif action.type_ != ActionType.comment:
            rules = self.state_rules[self.status]
            self.status = action.type_
            permissions = rules[action.type_]
            if not action.user.has_permission(permissions, self.division):
                raise ActionError(u"Insufficient permissions to perform that "
                                  u"action.")
        elif action.type_ == ActionType.comment:
            if action.user != self.submitter \
                    and not action.user.has_permission(PermissionType.elevated,
                            self.division):
                raise ActionError(u"You must either own or have special"
                                  u"privileges to comment on this request.")
        return action

    def __repr__(self):
        return "{x.__class__.__name__}({x.submitter}, {x.division}, {x.id})".\
                format(x=self)

    @db.validates('modifiers')
    def validate_add_modifier(self, attr, modifier):
        if self.status != ActionType.evaluating:
            raise ModifierError(u"Modifiers can only be added when the request"
                                u" is in an evaluating state.")
        if not modifier.user.has_permission(PermissionType.review,
                self.division):
            raise ModifierError(u"Only reviewers can add modifiers.")
        return modifier

    @property
    def transformed(self):
        """Get a special HTML representation of an attribute.

        Divisions can have a transformer defined on a for attributes that
        output a URL associated with that attribute. This property provides
        easy access to the output of any transformed attributes on this
        request.
        """
        class RequestTransformer(object):
            def __getattr__(self, attr):
                raw_value = getattr(self._request, attr)
                if attr in self._request.division.transformers:
                    transformer = self._request.division.transformers[attr]
                    return Markup(u'<a href="{link}">{value}</a>').format(
                            link=transformer(raw_value), value=str(raw_value))
                else:
                    return raw_value

            def __init__(self, request):
                self._request = request
        return RequestTransformer(self)



#!/usr/bin/env python
from setuptools import setup
import re


with open('evesrp/__init__.py', 'r') as f:
    init_contents = ''
    for line in f:
        init_contents += line + '\n'
version = re.search(r'^__version__ *= *u?[\'"]([^\'"]*)[\'"]', init_contents,
        re.MULTILINE)
if version:
    version = version.group(1)
else:
    raise Exception(u"Unable to find __version__ in evesrp/__init__.py")


setup(
    name=u'EVE-SRP',
    version=version,
    description=u'EVE Ship Replacement Program Helper',
    author=u'Will Ross',
    author_email=u'paxswill@paxswill.com',
    url=u'https://github.com/paxswill/evesrp',
    packages=[
        'evesrp',
        'evesrp.auth',
        'evesrp.views',
        'evesrp.util',
        'evesrp.migrate',
        'evesrp.migrate.versions',
    ],
    package_data={
        'evesrp': [
            'static/css/*.css',
            'static/css/*.css.map',
            'static/js/evesrp.min.js',
            'static/js/evesrp.min.js.map',
            'static/fonts/fontawesome-webfont.*',
            'static/ZeroClipboard.swf',
            'templates/*.html',
            'templates/*.xml',
            'migrate/alembic.ini',
            'migrate/script.py.mako',
        ],
    },
    classifiers=[
        u'Development Status :: 3 - Alpha',
        u'Framework :: Flask',
        u'License :: OSI Approved :: BSD License',
        u'Programming Language :: Python :: 3',
        u'Topic :: Games/Entertainment',
    ],
    dependency_links=[
        u'https://github.com/bravecollective/api/tarball/develop#egg=brave.api'
    ],
    install_requires=[
        'Flask==0.10.1',
        'Flask-Login==0.2.10',
        'Flask-Migrate==1.2.0',
        'Flask-Script==2.0.5',
        'Flask-SQLAlchemy==1.0',
        'Flask-WTF==0.9.4',
        'SQLAlchemy==0.9.3',
        'WTForms==1.0.5',
        'requests==2.2.1',
        'ecdsa==0.11',
        'six==1.7.3',
        'brave.api',
    ],
    entry_points={
        'console_scripts': [
            'evesrp = evesrp.util.manage:main',
        ],
    },
    zip_safe=False,
)



import six


def unistr(klass):
    def __str__(self):
        if six.PY2:
            return self.__unicode__().encode()
        else:
            return self.__unicode__()

    klass.__str__ = __str__
    return klass


def ensure_unicode(val):
    if isinstance(val, six.binary_type):
        return val.decode('utf-8')
    return val



"""empty message

Revision ID: 366b5d5a835
Revises: 3d35f833a24
Create Date: 2014-07-04 18:55:01.969566

"""

# revision identifiers, used by Alembic.
revision = '366b5d5a835'
down_revision = '3d35f833a24'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('modifier', '_type',
               existing_type=sa.VARCHAR(length=20),
               nullable=False)
    op.alter_column('transformerref', 'transformer',
               existing_type=postgresql.BYTEA(),
               nullable=True)
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('transformerref', 'transformer',
               existing_type=postgresql.BYTEA(),
               nullable=False)
    op.alter_column('modifier', '_type',
               existing_type=sa.VARCHAR(length=20),
               nullable=True)
    ### end Alembic commands ###



from __future__ import unicode_literals

"""This file was written by Dan Birken and discussed on his blog:
    http://www.danbirken.com/quicktip/2014/04/16/improve-developer-habits-with-db-diagnostisc.html

    File source: https://gist.github.com/danbirken/10939933
"""
import time
from sqlalchemy import engine
from sqlalchemy import event
 
class DbStats():
    def __init__(self):
        self.clear()
 
    def clear(self):
        self.total_queries = 0
        self.total_time = 0
 
    def add_query(self, timing):
        self.total_queries += 1
        self.total_time += timing
 
    def as_dict(self):
        return {
            'total_queries': self.total_queries,
            'total_time_ms': self.total_time * 1000,
        }
 
DB_STATS = DbStats()
 
@event.listens_for(engine.Engine, 'before_cursor_execute')
def before_cursor_execute(conn, cursor, statement, parameters,
                          context, exceutemany):
    context._query_start_time = time.time()
 
@event.listens_for(engine.Engine, 'after_cursor_execute')
def after_cursor_execute(conn, cursor, statement, parameters,
                         context, exceutemany):
    DB_STATS.add_query(time.time() - context._query_start_time)



"""Add API key table

Revision ID: 3d35f833a24
Revises: 4280bf2417c
Create Date: 2014-06-25 03:30:29.351874

"""

# revision identifiers, used by Alembic.
revision = '3d35f833a24'
down_revision = '4280bf2417c'

from alembic import op
import sqlalchemy as sa

def upgrade():
    # Add the table for API keys
    op.create_table('apikey',
            sa.Column('id', sa.Integer(), nullable=False),
            sa.Column('timestamp', sa.DateTime, nullable=False),
            sa.Column('user_id', sa.Integer(), nullable=False),
            sa.Column('key', sa.LargeBinary(length=32), nullable=False),
            sa.ForeignKeyConstraint(['user_id'], ['user.id']),
            sa.PrimaryKeyConstraint('id'))


def downgrade():
    op.drop_table('apikey')
